﻿namespace SelectionWinForms
{
    using System;
    using System.Collections.Generic;
    using System.Collections.ObjectModel;
    using System.Linq;
    using System.Windows;
    using System.Windows.Input;
    using System.Windows.Media;
    using System.Windows.Shapes;
    using Ptv.XServer.Controls.Map;
    using Ptv.XServer.Controls.Map.Canvases;

    public enum SelectMode
    {
        None,
        Click,
        Polygon
    }

    public enum SetMode
    {
        Set,
        Add,
        Xor
    }

    internal class SelectInteractor : WorldCanvas
    {
        public ObservableCollection<FrameworkElement> SelectedElements = new ObservableCollection<FrameworkElement>();

        private readonly MapView map;

        private readonly ObservableCollection<FrameworkElement> shapes;

        private System.Windows.Forms.Control control;
        private Polygon dragPolygon;

        private Point g1;
        private List<Point> polyPoints;

        private SelectMode selectMode = SelectMode.None;

        private WpfMap wpfmap;

        /// <summary>
        /// Initializes a new instance of the Select Interactor from WPF
        /// </summary>
        /// <param name="map"></param>
        /// <param name="shapes"></param>
        public SelectInteractor(MapView map, ObservableCollection<FrameworkElement> shapes) : base(map)
        {
            #region doc:SelectionPreparation

            this.map = map;
            this.shapes = shapes;

            map.MouseLeftButtonUp += this.map_MouseLeftButtonUp;
            map.MouseLeftButtonDown += this.map_MouseLeftButtonDown;
            map.MouseMove += this.map_MouseMove;

            SetZIndex(this, 999999999);
            this.SelectedElements.CollectionChanged += this.SelectedElements_CollectionChanged;

            #endregion doc:SelectionPreparation
        }

        /// <summary>
        /// Initializes a new instance of the SelectInteractor - to be used when used from WinForms
        /// </summary>
        /// <param name="map"></param>
        /// <param name="shapes"></param>
        /// <param name="wpfmap"></param>
        /// <param name="mapControl"></param>
        public SelectInteractor(MapView map, ObservableCollection<FrameworkElement> shapes, WpfMap wpfmap, System.Windows.Forms.Control mapControl) : base(map)
        {
            #region doc:SelectionPreparation

            this.map = map;
            this.shapes = shapes;
            this.wpfmap = wpfmap;
            this.control = mapControl;

            // use Preview events to prevent mouse events from being coonsumed elsewhere
            this.wpfmap.PreviewMouseLeftButtonDown += this.map_MouseLeftButtonDown;
            this.wpfmap.PreviewMouseLeftButtonUp += this.map_MouseLeftButtonUp;

            this.wpfmap.MouseMove += this.map_MouseMove;

            // override OnPaint handler - for testing
            this.control.Paint += this.MapControl_Paint;

            SetZIndex(this, 999999999);
            this.SelectedElements.CollectionChanged += this.SelectedElements_CollectionChanged;

            #endregion doc:SelectionPreparation
        }

        /// <summary>
        /// Helper function to get the Shape's parent of the given type T
        /// </summary>
        /// <typeparam name="T">the parent type to look for</typeparam>
        /// <param name="p_oElement">the element</param>
        /// <returns></returns>
        public static T GetLogicalParent<T>(DependencyObject p_oElement)
    where T : DependencyObject
        {
            DependencyObject oParent = p_oElement;
            Type oTargetType = typeof(T);

            if (p_oElement == null)
            {
                return null as T;
            }

            do
            {
                oParent = LogicalTreeHelper.GetParent(oParent);
            }
            while (!(oParent == null
                    || oParent.GetType() == oTargetType
                    || oParent.GetType().IsSubclassOf(oTargetType)));

            return oParent as T;
        }

        public void Remove()
        {
            this.map.MouseLeftButtonUp -= this.map_MouseLeftButtonUp;
            this.map.MouseLeftButtonDown -= this.map_MouseLeftButtonDown;
            this.map.MouseMove -= this.map_MouseMove;

            this.map.GeoCanvas.Children.Remove(this);
        }

        public override void Update(UpdateMode updateMode)
        {
            foreach (var e in Children.OfType<Ellipse>())
            {
                e.RenderTransform = new ScaleTransform(this.map.CurrentScale, this.map.CurrentScale);
            }
        }

        private void map_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            if (this.dragPolygon != null)
            {
                Children.Remove(this.dragPolygon);
            }

            this.g1 = CanvasToPtvMercator(e.GetPosition(this));

            if ((Keyboard.Modifiers & ModifierKeys.Alt) != 0)
            {
                this.selectMode = SelectMode.Polygon;
                this.polyPoints = new List<Point> { this.g1 };
                this.dragPolygon = new Polygon
                {
                    Fill = new SolidColorBrush(Color.FromArgb(0x3e, 0xa0, 0x00, 0x00)),
                    Stroke = new SolidColorBrush(Color.FromArgb(0x55, 0xff, 0x00, 0x00))
                };

                // this never shows up
                Children.Add(this.dragPolygon);
                e.Handled = true;
            }
            else
            {
                this.selectMode = SelectMode.Click;
            }
        }

        private void map_MouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            if ((Keyboard.Modifiers & ModifierKeys.Shift) == 0)
            {
                this.SelectedElements.Clear();
            }

            switch (this.selectMode)
            {
                case SelectMode.Click:
                    foreach (var shape in from shape in this.shapes let g2 = e.GetPosition(shape) let x = shape.InputHitTest(g2) let pin = GetLogicalParent<Ptv.XServer.Controls.Map.Symbols.Pin>(x as DependencyObject) where pin == shape select shape)
                    {
                        if (this.SelectedElements.Contains(shape))
                        {
                            this.SelectedElements.Remove(shape);
                        }
                        else
                        {
                            this.SelectedElements.Add(shape);
                        }
                    }

                    break;

                case SelectMode.Polygon:
                    {
                        CanvasToPtvMercator(e.GetPosition(this.wpfmap));

                        // mapControl.PanAndZoom.IsActive = true;
                        this.polyPoints.Add(this.polyPoints[0]);

                        // this wont work because without drawing the polygon because of the InputHitTest - eventually use GetLogicalParent like above
                        foreach (var shape in from shape in this.shapes let pp1 = (new Point(GetLeft(shape), GetTop(shape))) let x = this.dragPolygon.InputHitTest(pp1) where x == this.dragPolygon where !this.SelectedElements.Contains(shape) select shape)
                        {
                            this.SelectedElements.Add(shape);
                        }

                        Children.Remove(this.dragPolygon);
                        this.dragPolygon = null;
                    }

                    break;
            }

            this.selectMode = SelectMode.None;
        }

        private void map_MouseMove(object sender, System.Windows.Input.MouseEventArgs e)
        {
            if (this.selectMode != SelectMode.Polygon)
            {
                return;
            }

            Point tmPoint = e.GetPosition(this.wpfmap);
            this.polyPoints.Add(CanvasToPtvMercator(tmPoint));

            this.dragPolygon.Points.Clear();
            this.polyPoints.ForEach(point => this.dragPolygon.Points.Add(PtvMercatorToCanvas(point)));

            e.Handled = true;
        }

        private void MapControl_Paint(object sender, System.Windows.Forms.PaintEventArgs e)
        {
            if (this.SelectedElements.Count > 0)
            {
                foreach (FrameworkElement shape in this.SelectedElements)
                {
                    var p1 = new Point(GetLeft(shape), GetTop(shape));

                    var r = new Ellipse();
                    r.Tag = shape;
                    SetLeft(r, p1.X);
                    SetTop(r, p1.Y);
                    r.Width = r.Height = shape.ActualWidth * 2;
                    r.RenderTransform = new ScaleTransform(this.map.CurrentScale, this.map.CurrentScale);
                    r.RenderTransformOrigin = new Point(.5, .5);
                    r.Fill = new SolidColorBrush(Color.FromArgb(128, 255, 0, 0));

                    // this never shows up
                    Children.Add(r);

                    Color color = Colors.Aquamarine;
                    System.Drawing.Color newColor = System.Drawing.Color.FromArgb(color.A, color.R, color.G, color.B);

                    SolidColorBrush brush = new SolidColorBrush(Colors.Aquamarine);
                    System.Drawing.Pen pen = new System.Drawing.Pen(newColor, 2);
                }
            }
        }

        private void SelectedElements_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.NewItems != null)
            {
                foreach (FrameworkElement shape in e.NewItems)
                {
                    var r = new Ellipse();
                    r.Tag = shape;
                    Children.Add(r);

                    // change color of the Pin to indicate it was selected
                    (shape as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Red;
                }
            }

            if (e.OldItems != null)
            {
                var l = Children.Cast<object>().ToList();

                foreach (var selection in l.Cast<FrameworkElement>().Where(selection => e.OldItems.Contains(selection.Tag)))
                {
                    Children.Remove(selection);
                    (selection.Tag as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Blue;
                }
            }

            if (e.Action == System.Collections.Specialized.NotifyCollectionChangedAction.Reset)
            {
                var l = Children.Cast<object>().ToList();

                foreach (var selection in l.Cast<FrameworkElement>().Where(selection => selection.Tag != null))
                {
                    Children.Remove(selection);
                    (selection.Tag as Ptv.XServer.Controls.Map.Symbols.Pin).Color = Colors.Blue;
                }
            }
        }
    }
}